import os
import time
import json
import csv
import tempfile
import pdb

from em_tasks import logger
from em_tasks.utils.product_service import ProductService
from em_tasks.spree.api import SpreeApi


class ProductUtil():
  def __init__(self, endpoint, api_key, api_version='v1'):
    self.spree_api = SpreeApi(endpoint, api_key, api_version)
    # self.shop = self.spree_api.get_shop()
    # url = self.shop['url'] if self.shop['url'].startswith('http') else 'https://{}'.format(self.shop['url'])
    # parsed = urlparse(self.shop['url'])
    # self.shop_domain = parsed.hostname if parsed.hostname else 'everymarket.com'

  def get_amz_products(self, marketplace='us'):
    for inv in self.get_products('Amazon', marketplace):
      yield inv

  def get_ebay_products(self, marketplace='us'):
    for inv in self.get_products('Ebay', marketplace):
      yield inv

  def get_rakuten_products(self):
    for inv in self.get_products('Rakuten'):
      yield inv

  def get_naver_products(self):
    for inv in self.get_products('Naver'):
      yield inv

  def get_dangdang_products(self):
    for inv in self.get_products('Dangdang'):
      yield inv

  def get_coupang_products(self):
    for inv in self.get_products('Coupang'):
      yield inv

  def get_www_books_com_tw_products(self):
    for inv in self.get_products('www_books_com_tw'):
      yield inv

  def get_products(self, src, marketplace=None):
    source = None
    if src == 'Amazon':
      source = 'AMZ_{}'.format(marketplace.upper())
    elif src == 'Ebay':
      source = 'Ebay_{}'.format(marketplace.upper())
    elif src in ['Dangdang', 'Rakuten', 'Coupang', 'Naver', 'www_books_com_tw']:
      source = src

    if source is None:
      return

    _, output_path = tempfile.mkstemp('.csv', '{}-'.format(source))
    try:
      logger.info("[InventoryPath] %s", output_path)
      self.spree_api.download_inventory(source, output_path)

      if not os.path.isfile(output_path):
        return

      with open(output_path, encoding='utf-8', errors='ignore') as fh:
        reader = csv.reader(fh)
        headers = next(reader)
        headers_mapping = {
          'ProductID': 'product_id',
          'Source': 'source',
          'SourceProductID': 'source_product_id',
          'Handle': 'handle',
          'Variants': 'variants',
          'InStock': 'in_stock'
        }
        formatted_headers = []
        for header in headers:
          if header not in headers_mapping:
            continue

          formatted_headers.append(headers_mapping[header])

        for row in reader:
          if not row:
            continue

          record = dict()
          try:
            for idx, header in enumerate(formatted_headers):
              if header == 'variants':
                record[header] = json.loads(row[idx])
              else:
                record[header] = row[idx]
          except Exception as e:
            logger.info('[InventoryRow] %s', row)
            logger.exception(e)
            continue

          if not record['product_id'] or not record['source'] or not record['source_product_id']:
            continue

          yield record
    finally:
      if os.path.isfile(output_path):
        os.remove(output_path)

  def set_products_offer(self, prods, pool=None):
    store_offers = dict()
    for prod_id, prod in prods.items():
      store_offer = {'handle': prod['handle'], 'product_id': prod_id, 'offers': {}}
      variants = prod['variants']
      for variant in variants:
        store_offer['offers'][variant['sku']] = {
          'price': round(float(variant['price']), 2),
          'currency': variant['currency'],
          'quantity': int(variant['quantity']),
          'stock_items': variant['stock_items']
        }

      store_offers[prod_id] = store_offer

      offer = prod.get('offer', None)
      # Offer service encountered errors
      if isinstance(offer, bool) and not offer:
        continue

      # Product is not available on offer service
      if offer is None:
        continue

      # Product with multiple variants should supply each variant's offer
      if len(variants) > 1:
        if not isinstance(offer, dict):
          continue
      else:
        v = variants[0]
        if v['sku'] not in offer:
          offer = {v['sku']: offer}

      for v in variants:
        vid = v['variant_id']
        stock_items = v['stock_items']
        v_sku = v['sku']
        if v_sku not in offer:
          continue

        v_offer = offer[v_sku]
        if not v_offer:
          continue

        store_offer_v = store_offer['offers'][v_sku]

        try:
          src_offer = v_offer.get('src_offer', None)
          price = v_offer['price']
          quantity = v_offer['quantity']
          price_v = round(float(v['price']), 2)
          quantity_v = int(v['quantity'])
          if price <= 0 or quantity <= 0:
            if quantity_v <= 0:
              # Already out of stock
              logger.debug(
                '[OfferUnchanged] ProductId: %s, VariantId: %s, VariantPrice: %s, VariantQuantity: %s, Price: %s, Quantity: %s',
                prod_id, vid, price_v, quantity_v, price, quantity)
              continue

            stock_items_new = []
            for stock_item in stock_items:
              stock_item_new = dict(stock_item)
              stock_item_new['variant_id'] = vid
              stock_item_new['quantity'] = 0
              stock_items_new.append(stock_item_new)
            if pool:
              pool.apply_async(
                self.set_inventory,
                args=(stock_items_new,))
              logger.debug(
                '[VariantInventoryChanged] ProductId: %s, VariantId: %s, StockItems: %s',
                prod_id, vid, stock_items_new)
            else:
              succeed = self.set_inventory(stock_items_new)
              if succeed:
                logger.debug(
                  '[VariantInventoryChanged] ProductId: %s, VariantId: %s, StockItems: %s',
                  prod_id, vid, stock_items_new)
              else:
                logger.debug(
                  '[VariantInventoryUpdateFailed] ProductId: %s, VariantId: %s, StockItems: %s',
                  prod_id, vid, stock_items_new)

            store_offer_v['quantity'] = 0
          else:
            diff = abs(price_v - price)
            if src_offer and price > 0 and diff > 1:
              # No need to change price when price changes in 1 USD
              if pool:
                pool.apply_async(
                  self.set_price,
                  args=(prod_id, vid, price, src_offer.get('price', None),))
                logger.debug(
                  '[VariantPriceChanged] ProductId: %s, VariantId: %s, PriceNew: %s, PriceOld: %s, SrcOffer: %s',
                  prod_id, vid, price, price_v, src_offer)
              else:
                succeed = self.set_price(prod_id, vid, price, src_offer.get('price', None))
                if succeed:
                  logger.debug(
                    '[VariantPriceChanged] ProductId: %s, VariantId: %s, PriceNew: %s, PriceOld: %s, SrcOffer: %s',
                    prod_id, vid, price, price_v, src_offer)
                else:
                  logger.debug(
                    '[VariantPriceUpdateFailed] ProductId: %s, VariantId: %s, PriceNew: %s, PriceOld: %s, SrcOffer: %s',
                    prod_id, vid, price, price_v, src_offer)

              store_offer_v['price'] = price
            else:
              logger.debug(
                '[VariantPriceUnchanged] ProductId: %s, VariantId: %s, PriceNew: %s, PriceOld: %s, SrcOffer: %s',
                prod_id, vid, price, price_v, src_offer)

            if (quantity > 0 and quantity_v < 10) or (quantity <= 0 and quantity_v > 0):
              # No need to change quantity when quantity is still more than 10
              stock_items_new = []
              for stock_item in stock_items:
                stock_item_new = dict(stock_item)
                stock_item_new['variant_id'] = vid
                stock_item_new['quantity'] = quantity
                stock_items_new.append(stock_item_new)
              if pool:
                pool.apply_async(
                  self.set_inventory,
                  args=(stock_items_new,))
                logger.debug(
                  '[VariantInventoryChanged] ProductId: %s, VariantId: %s, StockItems: %s',
                  prod_id, vid, stock_items_new)
              else:
                succeed = self.set_inventory(stock_items_new)
                if succeed:
                  logger.debug(
                    '[VariantInventoryChanged] ProductId: %s, VariantId: %s, StockItems: %s',
                    prod_id, vid, stock_items_new)
                else:
                  logger.debug(
                    '[VariantInventoryUpdateFailed] ProductId: %s, VariantId: %s, StockItems: %s',
                    prod_id, vid, stock_items_new)
              store_offer_v['quantity'] = quantity
        except Exception as e:
          logger.exception(e)

    return store_offers

  def set_inventory(self, stock_items):
    succeed = True
    for stock_item in stock_items:
      max_retries = 3
      while max_retries > 0:
        try:
          resp = self.spree_api.update_stock(
            stock_item['id'], stock_item['quantity'], stock_item['stock_location_id'])
          if resp.status_code == 422:
            logger.debug(
              '[StockUpdateMsg] VariantId: %s, Stock: %s, Quantity: %s, Response: %s',
              stock_item['variant_id'], stock_item['id'], stock_item['quantity'], resp.text)
            max_retries -= 1
            time.sleep(1)

            continue

          try:
            stock = resp.json()
            if 'error' in stock or 'errors' in stock:
              succeed = False
              logger.warning(
                '[StockUpdateError] VariantId: %s, Stock: %s, Error: %s',
                stock_item['variant_id'], stock_item['id'],
                stock.get('error', stock.get('errors')))
            else:
              logger.info(
                '[StockUpdated] VariantId: %s, Stock: %s, Quantity: %s',
                stock_item['variant_id'], stock_item['id'], stock_item['quantity'])
          except Exception as e:
            succeed = False
            logger.exception(e)
            logger.warning('[StockUpdateInvalidResponse] %s', resp.text)

          break
        except Exception as e:
          logger.exception(e)
          max_retries -= 1
          time.sleep(1)

    return succeed

  def set_price(self, product_id, variant_id, price, cost=None):
    succeed = True
    max_retries = 3
    while max_retries > 0:
      try:
        if cost:
          resp = self.spree_api.update_variant(product_id, variant_id, price=price, cost_price=cost)
        else:
          resp = self.spree_api.update_variant(product_id, variant_id, price=price)
        if resp.status_code == 422:
          logger.debug(
            '[PriceUpdateMsg] ProductID: %s, VariantId: %s, Price: %s, Response: %s',
            product_id, variant_id, price, resp.text)
          max_retries -= 1
          time.sleep(1)

          continue

        sku = ''
        try:
          variant = resp.json()
          if 'error' in variant or 'errors' in variant:
            succeed = False
            logger.warning(
              '[PriceUpdateError] ProductId: %s, VariantId: %s, Error: %s',
              product_id, variant_id, variant.get('error', variant.get('errors')))
          elif 'sku' in variant and variant['sku']:
            sku = variant['sku']
            logger.info(
              '[PriceUpdated] ProductId: %s, VariantId: %s, Price: %s, SKU: %s',
              product_id, variant_id, price, sku)
          else:
            logger.info('[ResponseWithoutSKU] %s', resp.text)
        except:
          succeed = False
          logger.warning('[PriceUpdateInvalidResponse] %s', resp.text)

        break
      except Exception as e:
        succeed = False
        logger.warning(
          '[PriceUpdateError] ProductId: %s, VariantId: %s, Price: %s',
          product_id, variant_id, price)
        logger.exception(e)

    return succeed

  def download_inventory(self, source, output_path):
    self.spree_api.download_inventory(source, output_path)

    if not os.path.isfile(output_path):
      return

    with open(output_path, encoding='utf-8', errors='ignore') as fh:
      reader = csv.reader(fh)
      headers = next(reader)
      headers_mapping = {
        'ProductID': 'product_id',
        'Source': 'source',
        'SourceProductID': 'source_product_id',
        'Handle': 'handle',
        'Variants': 'variants'
      }
      formatted_headers = []
      for header in headers:
        if header not in headers_mapping:
          continue

        formatted_headers.append(headers_mapping[header])

      for row in reader:
        if not row:
          continue

        record = dict()
        try:
          for idx, header in enumerate(formatted_headers):
            if header == 'variants':
              record[header] = json.loads(row[idx])
            else:
              record[header] = row[idx]
        except Exception as e:
          logger.info('[InventoryRow] %s', row)
          logger.exception(e)
          continue

        yield record


if __name__ == "__main__":
  pass
